// Copyright (c) 2024 Dominic Masters
// 
// This software is released under the MIT License.
// https://opensource.org/licenses/MIT

#include "SceneLoader.hpp"
#include "game/Game.hpp"
#include "assert/assert.hpp"
#include "asset/loader/TextureLoader.hpp"
#include "scene/Scene.hpp"
#include "component/SceneComponentRegistry.hpp"

using namespace Dawn;

const std::string SceneLoader::ASSET_TYPE = "scene";

SceneLoader::SceneLoader(
  const std::shared_ptr<AssetManager> assetManager,
  const std::string name
) :
  AssetLoader(assetManager, name),
  state(SceneLoaderState::INITIAL)
{
}

void SceneLoader::setupDependencies() {
  std::cout << "Setting up dependencies" << std::endl;
  // Begin loading dependencies.
  auto &data = this->jsonLoader->data;
  if(data.contains("assets")) {
    for(auto &asset : data["assets"].items()) {
      auto &assetName = asset.key();
      auto &assetData = asset.value();
      assertTrue(assetData.contains("type"), "Asset missing type");
      assertTrue(assetData.contains("path"), "Asset missing path");
      auto type = assetData["type"].get<std::string>();
      auto path = assetData["path"].get<std::string>();

      std::shared_ptr<AssetLoader> loader;
      if(type == "texture") {
        loader = getAssetManager()->get<TextureLoader>(path);
      } else if(type == "json") {
        loader = getAssetManager()->get<JSONLoader>(path);
      } else {
        assertUnreachable("Unknown asset type: %s", type.c_str());
      }
      
      assertMapNotHasKey(
        ctx.assets,
        assetName,
        "Asset already exists: %s", assetName.c_str()
      );

      ctx.assets[assetName] = loader;
    }
  }
  this->state = SceneLoaderState::LOADING_DEPENDENCIES;
}

void SceneLoader::sceneInitializer(Scene &scene) {
  std::cout << "Initializing scene" << std::endl;

  auto &data = this->jsonLoader->data;
  if(data.contains("items")) {
    // Create the scene items
    for(auto &item : data["items"].items()) {
      auto &itemName = item.key();
      auto &itemData = item.value();
      auto sceneItem = scene.createSceneItem();
      ctx.items[itemName] = sceneItem;
    }

    // Add components to each scene item
    for(auto &item : data["items"].items()) {
      auto &itemName = item.key();
      auto &itemData = item.value();
      auto sceneItem = ctx.items[itemName];

      ctx.data = itemData;
      sceneItem->load(ctx);

      if(itemData.contains("components")) {
        for(auto &cmpItem : itemData["components"].items()) {
          auto &cmpName = cmpItem.key();
          auto &cmpData = cmpItem.value();
          assertTrue(cmpData.contains("type"), "Component missing type in %s", itemName.c_str());

          auto cmpType = cmpData["type"].get<std::string>();
          auto cmp = SceneComponentRegistry::createComponent(cmpType, sceneItem);
          ctx.data = cmpData;
          ctx.components[cmpName] = cmp;
          cmp->load(ctx);
        }
      }
    }
  }
  
  this->jsonLoader = nullptr;
  this->state = SceneLoaderState::DONE;
  getAssetManager()->remove(shared_from_this());
}

void SceneLoader::updateAsync() {
  switch(this->state) {
    case SceneLoaderState::INITIAL:
      this->jsonLoader = getAssetManager()->get<JSONLoader>(this->name);
      this->state = SceneLoaderState::LOADING_JSON;
      break;
    
    case SceneLoaderState::LOADING_JSON:
      assertNotNull(this->jsonLoader, "JSON Loader is NULL?");
      if(!this->jsonLoader->loaded) return;
      this->setupDependencies();
      break;

    case SceneLoaderState::LOADING_DEPENDENCIES:
      // Check if all dependencies are loaded.
      for(auto &asset : ctx.assets) {
        if(!asset.second->loaded) return;
      }

      this->state = SceneLoaderState::DEPENDENCIES_LOADED;
      break;

    case SceneLoaderState::DEPENDENCIES_LOADED:
      std::cout << "Deps Loaded" << std::endl;
      this->loaded = true;
      ctx.scene = std::make_shared<Scene>(
        this->getAssetManager()->getGame(),
        [this](Scene &scene) -> void {
          this->sceneInitializer(scene);
        }
      );
      this->state = SceneLoaderState::PENDING_STAGE;
      this->loaded = true;
      break;
    
    default:
      break;
  }
}

void SceneLoader::updateSync() {

}

std::string SceneLoader::getAssetType() const {
  return SceneLoader::ASSET_TYPE;
}

std::shared_ptr<Scene> SceneLoader::getScene() {
  return ctx.scene;
}

SceneLoader::~SceneLoader() {
  ctx = {};
  jsonLoader = nullptr;
  std::cout << "Scene Loader removed" << std::endl;
}