// Copyright (c) 2023 Dominic Masters
// 
// This software is released under the MIT License.
// https://opensource.org/licenses/MIT

#include "game/Game.hpp"
#include "scene/Scene.hpp"
#include "util/Flag.hpp"

#include "asset/loader/SceneLoader.hpp"

using namespace Dawn;

IGame::IGame() {

}

void IGame::init() {
  assertFlagOff(state, GAME_STATE_INITIALIZED, "Game already initialized?");
  Flag::turnOn<uint8_t>(state, GAME_STATE_INITIALIZED);

  auto selfAsGame = this->getSelfAsGame();

  renderHost = std::make_shared<RenderHost>();
  renderHost->init(selfAsGame);
  
  assetManager = std::make_shared<AssetManager>();
  assetManager->init(selfAsGame);
  
  localeManager = std::make_shared<LocaleManager>();
  localeManager->init(selfAsGame);

  #ifdef DAWN_ENABLE_PHYSICS
    physicsManager = std::make_shared<PhysicsManager>();
    physicsManager->init(selfAsGame);
  #endif

  inputManager.init(selfAsGame);
  saveManager.init(selfAsGame);

  this->initManagers();

  // TEST
  auto scene = this->assetManager->get<SceneLoader>("test_scene.json");
  while(!this->assetManager->isEverythingLoaded()) {
    this->assetManager->update();
  }

  // auto initialScene = this->getInitialScene();
  // nextFrameScene = std::make_shared<Scene>(selfAsGame, initialScene);
  nextFrameScene = scene->getScene();
}

void IGame::deinit() {
  assertFlagOn(state, GAME_STATE_INITIALIZED, "Game not initialized?");

  if(currentScene) currentScene->deinit();
  currentScene = nullptr;
  
  if(nextFrameScene) nextFrameScene->deinit();
  nextFrameScene = nullptr;

  #ifdef DAWN_ENABLE_PHYSICS
    physicsManager = nullptr;
  #endif

  assetManager = nullptr;
  renderHost = nullptr;

  Flag::turnOff<uint8_t>(state, GAME_STATE_INITIALIZED);
}

void IGame::update() {
  this->assetManager->update();
  this->inputManager.update();

  if(nextFrameScene) {
    if(currentScene) currentScene->deinit();
    nextFrameScene->init();
    currentScene = nextFrameScene;
    nextFrameScene = nullptr;
  }

  timeManager.update();
  if(currentScene) currentScene->update();

  #ifdef DAWN_ENABLE_PHYSICS
    physicsManager->update();
  #endif
  
  renderHost->update(this->getSelfAsGame());
}

bool_t IGame::isCloseRequested() {
  return (
    renderHost->isCloseRequested()
  );
}

std::shared_ptr<Scene> IGame::getCurrentScene() {
  return currentScene;
}

std::shared_ptr<Game> IGame::getSelfAsGame() {
  return std::static_pointer_cast<Game>(shared_from_this());
}

IGame::~IGame() {
  assertFlagOff(state, GAME_STATE_INITIALIZED, "Game not deinited properly?");
}