// Copyright (c) 2023 Dominic Masters
// 
// This software is released under the MIT License.
// https://opensource.org/licenses/MIT

#include "assert/assert.hpp"
#include "util/Flag.hpp"
#include "game/Game.hpp"
#include "scene/Scene.hpp"

using namespace Dawn;

void SceneComponent::init(const std::shared_ptr<SceneItem> item) {
  assertFlagOff(
    sceneComponentState,
    SCENE_COMPONENT_STATE_INIT,
    "SceneComponent is already initialized!"
  );
  Flag::turnOn<uint_fast8_t>(
    sceneComponentState,
    SCENE_COMPONENT_STATE_INIT
  );
  this->events.clear();
  this->item = item;
  this->onInit();
}

void SceneComponent::dispose() {
  assertFlagOn(
    sceneComponentState,
    SCENE_COMPONENT_STATE_INIT,
    "SceneComponent is not initialized!"
  );
  assertFlagOff(
    sceneComponentState,
    SCENE_COMPONENT_STATE_DISPOSED,
    "SceneComponent is already disposed!"
  );
  Flag::turnOn<uint_fast8_t>(
    sceneComponentState,
    SCENE_COMPONENT_STATE_DISPOSED
  );

  for(auto &event : this->events) {
    event();
  }
  this->events.clear();
  this->onDispose();
  this->item.reset();
}

bool_t SceneComponent::isInitialized() {
  return Flag::isOn<uint_fast8_t>(
    sceneComponentState,
    SCENE_COMPONENT_STATE_INIT
  );
}

std::shared_ptr<SceneItem> SceneComponent::getItem() {
  auto item = this->item.lock();
  assertNotNull(item, "SceneItem has unloaded?");
  return item;
}

std::shared_ptr<Scene> SceneComponent::getScene() {
  return this->getItem()->getScene();
}

std::shared_ptr<Game> SceneComponent::getGame() {
  return this->getScene()->getGame();
}

void SceneComponent::load(const SceneComponentLoadContext &context) {
  // Override this method to load data from a JSON object.
}

SceneComponent::~SceneComponent() {
  if(Flag::isOn<uint_fast8_t>(
    sceneComponentState,
    SCENE_COMPONENT_STATE_INIT
  )) {
    assertFlagOn(
      sceneComponentState,
      SCENE_COMPONENT_STATE_DISPOSED,
      "SceneComponent is initialized but was not properly disposed!"
    );
  }
}